/* Programme Démonstrateur directement issu de la version M pour ajouter l'option de référence
   d'orientation Lacet en CAP magnétique.
   Programme à associer avec P22N_Demonstrateur_Sonde.ino
   
   ATTENTION : Pour que les textes affichés soient cohérents, il faut impérativement utiliser
               P35_Ecrire_les_textes_en_EEPROM.ino pour les inscrire en EEPROM.
   
   UTILISATION DU PROGRAMME :
 ----------------------------
 1) Établir la liaison TX/RX avec l'inverseur,
 2) Faire un RESET sur la Raquette,
 3) Tester toutes les fonctions relatives à l'option de référence Gyroscopique / Cap magnétique.
  
  Ce que fait ce démonstrateur :
• Mise en place de l'inversion de l'option G/C avec le BP central du codeur rotatif.
• Imposer la valeur de la consigne par un message de type "Cnnn*).
• Positionne le curseur. (Touches S6 et S14)
• Traite et transmet la consigne de CAP magnétique de référence.
• Pour simplifier le lecture de la boucle de base le gros du traitement est placé en subroutines.
• Gère le calcul de l'écart magnétique et affiche les données et retour à la gestion par interruption
  du codeur rotatif incrémental.

                                                             PGM            RAM dynamique
Démonstrateur sans modification :                      24704 Octets (80%)   445 Octets (21%)
Mise en place de l'inversion de l'option G/C :         24784 Octets (80%)   446 Octets (21%)   +80 octets
Remplacé Chaine_Memorisee == "OK" par ACR_egal_OK() :  24776 Octets (80%)   446 Octets (21%)    -8 octets
Ecran de saisie et imposer la valeur de la consigne :  25068 Octets (81%)   446 Octets (21%)  +292 octets
Positionne le curseur. (Touches S6 et S14) :           25174 Octets (81%)   450 Octets (21%)  +106 octets
Traite et transmet la consigne de CAP magnétique :     25428 Octets (82%)   450 Octets (21%)  +254 octets
Boucle de base simplifiée en appels de procédures :    25452 Octets (82%)   450 Octets (21%)   +24 octets
Calcule l'écart M et affichage et Interruptions :      25628 Octets (83%)   450 Octets (21%)  +176 octets

//--------------------- Déclarations pour le matériel ----------------------- */
#include <EEPROM.h>
#include <Adafruit_ssd1306syp.h> // Afficher graphique OLED.
// Déclarations pour le capteur codeur rotatif KY-40 :
#define Codeur_A 3    // D2 : DT  >>>>>>>>>>>>>>>>>>> ATTENTION : En fonction du codeur KY-40 approvisionné il
#define Codeur_B 2    // D3 : CLK >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> faudra éventuellement inverser D2 et D3 pour
// Définition des broches pour la ligne I2C.   >>>>>>>>>>>>>>>>>> obtenir un sens de variation correct.
#define broche_SDA 11 // D11 utilisée en sortie.
#define broche_SCL 12 // D12 utilisée en sortie.
// Initialisation des broches en sorties et affectations.
Adafruit_ssd1306syp display(broche_SDA, broche_SCL);
//--------------- Définition des broches d'Entrées Analogiques --------------
#define Mesure_U_clavier 6 // Entrée analogique A6 utilisée pour le clavier.
#define BP_central 21 // A7 pour lire le bouton central.
//------------------ Sorties sur broches "Analogiques" ----------------------
#define LED_Apprentissage 14 // Sortie A0 pilote la LED "JEKERT est en mode Apprentissage.
#define LED_Stabilisation_Gyroscopique 15 // Sortie A1 pilote la LED JAUNE de "stabilisation gyroscopique active".
#define LED_SECURITE 16 // Sortie A2 pilote la LED ROUGE de SECURITÉ.
#define LED_MOUVEMENTS 17 // Sortie A3 pilote la LED BLEUE : MOUVEMENTS.
#define LED_Retour_par_FIN 18 // Sortie A4 pilote la LED Retour par FIN.
#define LED_Clavier 19 // Sortie A5 pilote la LED Clavier.
//-------------- Définition des broches de sorties Binaires -----------------
#define LED_Erreur 4 // LED rouge : MOUVEMENTS ou ERREUR.
#define LED_mode_GRAPHIQUE 5 // LED mode GRAPHIQUE.
#define BUZZER 6
#define C1 7  // Colonne C1 sur la sortie D7.
#define C2 8  // Colonne C1 sur la sortie D8.
#define C3 9  // Colonne C1 sur la sortie D9.
#define C4 10 // Colonne C1 sur la sortie D10.
#define LED_Loop_active 13 // D13 utilisée pour la LED clignotante "Boucle de base".
//------------------------ Constantes du programme --------------------------
#define Version_PGM "23-02-2018"    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_donnees 1       //@@@@@@@@ (1 en version définitive.) @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_options 5       //@@@@@@@@ (5 en version définitive.) @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_Exploitation 4  //@@@@@@@@ (4 en version définitive.) @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_des_postures 9  //@@@@@@@@ (9 en version définitive.) @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#define Num_initial_Apprentissage 9 //@@@@@@@@ (9 en version définitive.) @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

#define LGR_chaine 20
//------------------------- Variables du programme --------------------------
// Note : Les déclarations sont classées par type et par ordre alphabétique croissant.
boolean A_set, B_set;   // État des sorties déphasées du codeur.
boolean Affichage_NAV_Continu;
boolean Change_le_chainage_PGM;
boolean Change_le_num_PGM_cible;
boolean Gradateur; // Si true le Bouton rotatif agit sur le gradateur, si false agit sur les Items des menus.
boolean INT_active;     // Indique un INT_active d'état détecté.
boolean LED_Erreur_clignote;
boolean LED_Retour_par_FIN_clignote;
boolean LASER; // Allumer/éteindre.
boolean Mode_Graphique;
boolean Mode_MANUEL;
boolean Mode_pointage_LASER;
boolean Mode_saisie_de_CAP_magnetique;
boolean PHARES;
volatile boolean Rotation;  // Indique une activation du codeur rotatif.
boolean Reference_gyroscopique;
boolean Sens_positif; // Sens de rotation.
boolean Sentinelle_detectee; // Délimiteur de fin détecté.
boolean Verifier_OK; // Vérifier que l'ACR est bien "OK*".
boolean Vertical; // Le LASER est orienté verticalement. (Consigne potentiométrique.)

byte Articulation;
byte BP; // Bouton du claviar activé.
byte BIT_Option;
byte Code_Consigne;
byte Compteur_caracteres;
byte Mode_en_cours; // 1 : EXPLOITER.   2 : Imposer des POSTURES. (Correspondent au BP du clavier.)
                    // 3 : OPTIONS de la sonde.   4 : DONNEES de la sonde.    5: MOUVOIR.
                    // 6 : Mode APPRENTISSAGE.
byte NB_a_incrementer_decrementer;
byte NB_PGM_a_chainer;
byte NB_Repetition_Mvt;
byte Num_des_donnees;
byte Num_Exploitation;
byte Num_des_options;
byte Num_des_postures;
byte Num_Apprentissage;
byte Num_PGM;
byte Poids_decimal;
byte Position_aff_CAP;
byte RETOUR; // Code de retour d'une fonction à risque. (Code 0 : pas de retour nécessaire.)
byte SAV_Num_PGM;
byte SAVX, SAVY; // Anciennes valeurs pour effacer.
byte X,Y; // Coordonnées de l'extrémité du rayon tracé.

char Jambe;
char Caractere;

int Angle;
int Calage_Gyroscopique;
int Cap_magnetique_de_consigne;
int CNA; // Mesure analogique retournée par le CNA.
int Ecart_gyroscopique;
int Roulis, Ruban, Tangage;

unsigned long Ancienne_valeur_Temps_Ecoule_depuis_RESET;

float Alpha;

String Chaine_Memorisee; // Tampon mémoire pour la chaine entrante.

void setup()  {
  //---------------------- Définition des vecteurs d'interruption ----------------------
  attachInterrupt(0, Front_sur_A, CHANGE); // Appel à Front_sur_A si INT_active d'état détecté.
  // Broche de l'interruption 1 (E/S D3) :
  attachInterrupt(1, Front_sur_B, CHANGE); // Appel à Front_sur_B si INT_active d'état détecté.
  //------------------------ Initialisations pour le matériel --------------------------
  display.initialize(); delay(300); display.clear();
  Serial.begin(115200);
  //------------------------------ Affiche écran initial -------------------------------
  display.setTextSize(1); // Taille la plus petite.
  Affiche_cadre_du_haut(11); Aff_TEXTE_EEPROM(202,8); display.print(Version_PGM);
  CADRE(); display.setTextSize(2); // Taille double.
  display.setCursor(38,24); Aff_TEXTE_EEPROM(0,5);
  display.setCursor(32,41); AFF_mot_JEKERT();
  display.update(); display.setTextSize(1); // Taille la plus petite.
  //-------------------------- Initialisations des booléens ----------------------------
  PHARES = false; LASER = false; Gradateur = false; RETOUR = 0; LED_Retour_par_FIN_clignote = false;
  Mode_pointage_LASER = false; Vertical = true; LED_Erreur_clignote = false; Mode_MANUEL = false;
  Change_le_num_PGM_cible = false; NB_PGM_a_chainer = 1; Change_le_chainage_PGM = false;
  Mode_Graphique = false; Reference_gyroscopique = false; Mode_saisie_de_CAP_magnetique = false;
  Affichage_NAV_Continu = false;
  //------------------------- Initialisations des variables ----------------------------
  NB_Repetition_Mvt = 1; Jambe = 'A'; Articulation = 1; 
  //------------------------------ Initialisations des E/S -----------------------------
  pinMode(C1,OUTPUT); pinMode(C2,OUTPUT); pinMode(C3,OUTPUT); pinMode(C4,OUTPUT); digitalWrite(C1,LOW);
  digitalWrite(C2,LOW); digitalWrite(C3,LOW); digitalWrite(C4,LOW); pinMode(LED_Clavier,OUTPUT); 
  pinMode(LED_Erreur,OUTPUT); digitalWrite(LED_Erreur,LOW); pinMode(Codeur_A,INPUT); pinMode(Codeur_B,INPUT);
  pinMode(LED_Retour_par_FIN,OUTPUT); digitalWrite(LED_Retour_par_FIN,LOW);
  pinMode(LED_MOUVEMENTS,OUTPUT); digitalWrite(LED_MOUVEMENTS,LOW);
  pinMode(LED_mode_GRAPHIQUE,OUTPUT); digitalWrite(LED_mode_GRAPHIQUE,LOW);
  pinMode(LED_SECURITE,OUTPUT); Eteindre_SECURITE();
  pinMode(LED_Stabilisation_Gyroscopique,OUTPUT); digitalWrite(LED_Stabilisation_Gyroscopique,LOW);
  pinMode(LED_Apprentissage,OUTPUT); Eteindre_Apprentissage();  
  //------------------------------------------------------------------------------------
  Chaine_Memorisee.reserve(LGR_chaine); // Reserve LGR_chaine octets pour la chaine.
  //ATTENDRE_un_BP();  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  Mode_en_cours = 1; // MENU EXPLOITER pour faire sortir la sonde de l'hibernation ...
  Num_Exploitation = Num_initial_Exploitation; Affiche_Ecran_MENU(); // Pour afficher l'item "Réveiller JEKERT". 
  Rotation = true ; Sens_positif = true;} // LOW initial simule une rotation pour déclencher le premier menu.

void loop() {
  //------------- Faire clignoter la LED "boucle de base" et Mouvements ----------------
  if (millis() > Ancienne_valeur_Temps_Ecoule_depuis_RESET) {
      Ancienne_valeur_Temps_Ecoule_depuis_RESET = millis() + 100; // Réarme pour 0,1S ou 0,03S.
      digitalWrite(LED_Loop_active,!digitalRead(LED_Loop_active)); // Inverser l'état de la LED.
      if (LED_Erreur_clignote) digitalWrite(LED_Erreur,!digitalRead(LED_Erreur));
      else digitalWrite(LED_Erreur,LOW);
      if (LED_Retour_par_FIN_clignote) digitalWrite(LED_Retour_par_FIN,!digitalRead(LED_Retour_par_FIN));
      else digitalWrite(LED_Retour_par_FIN,LOW);}
  //------------------------------- Traiter le clavier ---------------------------------
  TRAITER_LE_CLAVIER();
  //----------------------------- Traite le B.P. central -------------------------------
  TRAITE_LE_BP_CENTRAL_DU_CODEUR_ROTATIF();
  //--------------------- Traite la rotation du codeur incrémental ---------------------
  TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL();
  //----------------- Traite l'affichage continu des données de NAVIGATION -------------
  if (Affichage_NAV_Continu) Rafraichir_les_donnees_de_navigation();}
  //================================================================================================

void TRAITE_LA_ROTATION_DU_CODEUR_INCREMENTAL() {
  INT_active = true; // Armer un nouveau comptage.
  if ((Rotation) && (!Affichage_NAV_Continu)) {Rotation = false;
      if (Mode_saisie_de_CAP_magnetique) {Traite_saisie_cap_magnerique_de_reference();}
      else {Efface_le_code(); Effacer_ACR(); Afficher_Rot_ou_CLV(); // Pour effacer un éventuel message d'erreur.
            //-------------------------------------------------------------------------------
            if (Gradateur) Traite_Gradateur(); // Si NON : Traite les menus ci-dessous.
            else {if (Mode_en_cours == 1) Effacer_zone(6,55,17); // Effacer "Retour par FIN" si mode EXPLOITER.
                  if (Mode_en_cours == 5) {// Menu MOUVOIR.
                      NB_a_incrementer_decrementer = NB_Repetition_Mvt; Incrementer_Decrementer(9);
                      NB_Repetition_Mvt = NB_a_incrementer_decrementer;
                      Afficher_le_NB_fois();}           
                  Traite_Item_Menus(); display.update();}}}}
                  
//-------------------------------------- Traitement du CLAVIER -------------------------------------
void TRAITE_LE_BP_CENTRAL_DU_CODEUR_ROTATIF() {
  if (analogRead(BP_central) < 512) { // Si le B.P. central est activé :
      digitalWrite(LED_Clavier,HIGH);
      while (analogRead(BP_central) < 512); // Attendre le relâché.
      digitalWrite(LED_Clavier,LOW);
      //---------------------------------------------------
      if (Mode_en_cours == 1) // Mode EXPLOITATION.
          switch (Num_Exploitation) {
            case 11  : {Envoyer_Consigne(66); break;} // Inverser l'amplitude du débattements si Exploitation LASER.
            case 15  : {Envoyer_Consigne(66); break;} // Inverser l'amplitude du débattements si pilotage Manuel.
             default : Envoyer_Consigne(90);} // Générer une temporisation d'une seconde.
      if (Mode_en_cours == 4) {Mode_Graphique = !Mode_Graphique; // Mode DONNEES.
                               digitalWrite(LED_mode_GRAPHIQUE,Mode_Graphique);}
      if ((Mode_en_cours == 2) || (Mode_en_cours == 5)) Envoyer_Consigne(14); // Libérer les efforts.
      if (Mode_en_cours == 3) {
          if (Affichage_NAV_Continu) {
              FIN_affichage_en_continu(); Reference_gyroscopique = !Reference_gyroscopique;
              if (Reference_gyroscopique) Envoyer_Consigne(98); else Envoyer_Consigne(97);
              Affiche_ecran_graphique(); Ouvre_AFF_en_continu();}
          else Inverse_LED_SECURITE();}     
      if (Mode_en_cours == 6) Inverse_LED_SECURITE();}}
      
void TRAITER_LE_CLAVIER() { 
  Tester_le_clavier();
  if (BP > 0) { // Si BP > 0 c'est qu'un B.P. a été activé.
      Effacer_ACR();
      if (Mode_saisie_de_CAP_magnetique) {
          switch (BP) {
            case  6 : {Poids_decimal = Poids_decimal / 10; if (Poids_decimal == 0) Poids_decimal = 100;
                       Affiche_le_curseur(); break;}           
            case  8 : {Fermer_la_saisie_du_CAP_magnetique(); break;}           
            case 14 : {Poids_decimal = Poids_decimal * 10; if (Poids_decimal > 100) Poids_decimal = 1;
                       Affiche_le_curseur(); break;}
            default : BIP();}}
      //================================================================================
      else {if (Mode_pointage_LASER) {
                if ((BP == 9) || (BP == 11)) Envoyer_Consigne(64); // Balayage vertical.
                if ((BP == 6) || (BP == 14)) Envoyer_Consigne(65); // Balayage horizontal.
                if (BP == 8) {Envoyer_Consigne(63); Purge_les_booleens();} // Touche FIN pour sortir.
                if (BP == 10) Envoyer_Consigne(9);} // Simuler un tir au LASER.
           // NOTE : L'exploration de S15 n'est pas effectuée dans cette séquence car réalisée dans
           //        la séquence "else de balayage des touches de MENU". Il y aurait contradiction ici.
           //================================================================================
           if (Mode_MANUEL) {
               switch (BP) {
                 case  5 : {Envoyer_Consigne(79); Articulation = 1; break;}
                 case  8 : {Traite_la_touche_NON_ou_FIN(); Affiche_Ecran_MENU(); Curseur_1_37();
                            AFF_TXT_EEPROM_et_RETOUR(691,17); display.update(); break;}
                 case  9 : {Envoyer_Consigne(78); Jambe = 'D'; break;}
                 case 11 : {Envoyer_Consigne(77); Jambe = 'C'; break;}
                 case 12 : {Envoyer_Consigne(80); Articulation = 2; break;}
                 case 13 : {Envoyer_Consigne(76); Jambe = 'B'; break;}
                 case 15 : {Envoyer_Consigne(81); Articulation = 3; break;}
                 case 16 : {Envoyer_Consigne(75); Jambe = 'A';}}
                 if (Gradateur) Affiche_moteur_actif();}
           //================================================================================
           else {//---------------------- Touches de changement de MENU. ----------------------------------------
                 if (BP < 6) {Mode_en_cours = BP; Sens_positif = true; // Pour ouvrir sur la bonne page écran.
                              Num_des_donnees = Num_initial_donnees; Num_des_postures = Num_initial_des_postures;
                              Num_des_options = Num_initial_options; Num_Exploitation = Num_initial_Exploitation; 
                              Affiche_Ecran_MENU(); Traite_Item_Menus();} // Boutons à validité permanente.
                 if (BP == 10) {Num_Apprentissage = Num_initial_Apprentissage;
                                Sens_positif = true; MENU_APPRENTISSAGE();}
           //----------------------------------------------------------------------------------------------
           if (BP == 7) Traite_la_touche_OUI();
           if (BP == 8) Traite_la_touche_NON_ou_FIN();
           Teste_Touche_PHARES (); // Traite la touche 13.
            if (!Mode_MANUEL) {Teste_Touche_LASER();} // Teste la touche 15 et allume ou éteind le LASER.
           //------------------------------------- Traite les MOUVEMENTS ----------------------------------
           if (Mode_en_cours == 5) {// Mode MOUVOIR
               for (byte I = NB_Repetition_Mvt; I>0; I--)
                   switch (BP) { // Traite les touches de déplacement.
                     case 6  : {Envoyer_Consigne(7); break;} // Translation à droite.
                     case 9  : {Envoyer_Consigne(3); break;} // Avancer.
                     case 11 : {Envoyer_Consigne(4); break;} // Reculer.
                     case 12 : {Envoyer_Consigne(5); break;} // Tourner à droite.
                     case 14 : {Envoyer_Consigne(8); break;} // Translater à gauche.
                     case 16 :  Envoyer_Consigne(6);}}}}}}   // Tourner à gauche.

void Traite_la_touche_OUI() {
  //-------------------------------------------------------------------------------
  if (Mode_en_cours == 1) { // Menu EXPLOITER.
      switch (Num_Exploitation) {
        case  1 : {Envoyer_Consigne(69); Affiche_le_balayage_ultrasons(); break;} // Enregistrer balayage ultrasons.     
        case  2 : {Affiche_le_balayage_ultrasons(); break;} //Afficher la télémétrie.
        case  3 : {Envoyer_Consigne(18); AFF_POSTURE(83); break;} // Enregistrer en EEPROM la posture actuelle.
        case  4 : {Envoyer_Consigne(61); Eteindre_SECURITE(); Affichage_NAV_Continu = false;
                   LASER = false; break;}  // Passer la Sonde en Hibernation. (QUITTER.)
        case  5 : {Envoyer_Consigne(55); break;} // Réveiller la sonde.
        case  6 : {Envoyer_Consigne(58); break;} // Test de communication MPU6050.
        case  7 : {Envoyer_Consigne(59); Affiche_Calage_Gyroscope(); break;} // Calage du gyroscope.
        case  8 : {Envoyer_Consigne(54); break;} // LASER et Phares forcés à 127.
        // Signale à la sonde que le gradateur agit sur les phares :
        case  9 : {Envoyer_Consigne(43); Gradateur = true; AFFICHE_puissance_eclairages();
                   LED_Retour_par_FIN_clignote = true; break;}
        // Signale à la sonde que le gradateur agit sur le LASER :
        case 10 : {Envoyer_Consigne(42); Gradateur = true; AFFICHE_puissance_eclairages();
                   LED_Retour_par_FIN_clignote = true; break;}
        case 11 : {Envoyer_Consigne(62); Mode_pointage_LASER = true; LED_Retour_par_FIN_clignote = true;
                   LED_Erreur_clignote = true; Gradateur = true; LASER = false; break;}
        case 12 : {Envoyer_Consigne(67); AFF_MENU_Spectre_colorimetrique(); break;} // Enregistrer colorimétrique.
        case 13 : {AFF_MENU_Spectre_colorimetrique(); break;} // Visualiser un spectre colorimétrique.
        case 14 : {LED_Retour_par_FIN_clignote = true; LED_Erreur_clignote = true; // Mvt Torsion. 
                   Envoyer_Consigne(38); RETOUR = 39; Gradateur = true; break;} // Pour BR envoie (40) & (41).   
        case 15 : {LED_Retour_par_FIN_clignote = true; LED_Erreur_clignote = true; // Pilotage MANUEL. 
                   Envoyer_Consigne(73); RETOUR = 74; Gradateur = true; Mode_MANUEL = true;
                   Affiche_moteur_actif(); Envoyer_Consigne(79); Envoyer_Consigne(75); break;}
        case 16 : {AFF_POSTURE(82); break;} // Aff. posture Actuelle.
        case 17 : {AFF_POSTURE(83); break;} // Aff. posture SAV en EEPROM.
        case 18 : {Envoyer_Consigne(9); break;} // Tir LASER.
        case 19 : Saisie_du_CAP_magnetique();}}
  //-------------------------------------------------------------------------------
  if (Mode_en_cours == 2)  { // Menu des POSTURES. 
      switch (Num_des_postures) {
        case 1 : {Envoyer_Consigne(1); break;}  // Posture VEILLE. (Sonde posée au sol.)
        case 2 : {Envoyer_Consigne(16); RETOUR = 17; LED_Retour_par_FIN_clignote = true; break;} // STBRaisonnable.
        case 3 : {Envoyer_Consigne(10); LED_Retour_par_FIN_clignote = true; RETOUR = 11; break;} // Hauteur MAXI.
        case 4 : {Envoyer_Consigne(15); break;} // Moteurs au Neutre opérationnel.
        case 5 : {Envoyer_Consigne(19); break;} // Imposer la posture apprise en EEPROM.
        case 6 : {Envoyer_Consigne(23); break;} // Configuration décollage.
        case 7 : {Envoyer_Consigne(22); break;} // Atterrissage.
        case 8 : {Envoyer_Consigne(21); break;} // Déposer la sonde.
        case 9 : Envoyer_Consigne(2);}}         // Stable Transversal.    
  //-------------------------------------------------------------------------------
  if (Mode_en_cours == 3) { // Menu des OPTIONS.
    switch (Num_des_options) { // Change l'étét du booléen système.
      case 1 : {Envoyer_Consigne(25); break;} // Disjoncter les énergies.
      case 2 : {Envoyer_Consigne(32); break;} // Moteurs ON.
      case 3 : {Envoyer_Consigne(33); break;} // Mouvements rapides.
      case 4 : {if (digitalRead(LED_SECURITE)) 
                   {Envoyer_Consigne(35); Positionne_LED_Stab_Gyroscopique(); Eteindre_SECURITE();}
                else ERR_NON_Securite(); break;}  // Stabilisation Gyroscopique.
      case 5 : {if (digitalRead(LED_SECURITE)) Ouvre_AFF_en_continu();
                else ERR_NON_Securite(); break;}} // Affichage en continu des données.
      if (!Affichage_NAV_Continu) Affiche_etat_option_en_cours();}
  //-------------------------------------------------------------------------------
  if (Mode_en_cours == 6) { // Menu APPRENTISSAGE.
      switch (Num_Apprentissage) {
        case 1 : {Demande_Information(94); // Liste le PGM indexé en EEPROM si NON vide.
                  if (Chaine_Memorisee == "OK") Lister_un_PGM(); else Chaine_Memorisee = "E15";
                  AFF_ACR(); break;}
        case 2 : {Effacer_le_dernier_code_appris(); break;}
        case 3 : {Effacer_un_PGM(); break;} // Efface le PGM indexé en EEPROM.
        case 4 : {Envoyer_Consigne(12); break;} // Le PGM indexé devient la SOURCE d'un transfert.
        case 5 : {Transferer_la_SOURCE(); break;} // La SOURCE passe dans le PGM actuel et est effacée.
        case 6 : {if (digitalRead(LED_SECURITE))
                      {if (!Apprentissage_en_cour()) Enchainer_les_PGMs();
                       else {BIP_et_XY(); display.print("E19");}
                       Eteindre_SECURITE();}
                  else ERR_NON_Securite(); break;}
        case 7 : {if (digitalRead(LED_SECURITE)) {Chainer_des_PGM(); AFF_ACR();}
                  else ERR_NON_Securite(); break;} // Imposer un nombre de PGM chaînés.
        case 8 : {Envoyer_Consigne(71); BIT_Option = 7; // Mode APPRENTISSAGE sur JEKERT : OUI.
                  digitalWrite(LED_Apprentissage,HIGH); Affiche_etat_option_en_cours(); break;}        
        case 9 : {Changer_de_PGM_EEPROM(); AFF_ACR(); break;}}} // Change le n° du PGM cible.
  display.update();}

void Traite_la_touche_NON_ou_FIN() { // La touche NON est aussi la touche FIN.
  Purge_les_booleens(); // Pour que le BR agisse sur les MENUs.
  //------------------------------------------------------------------------------- 
  if (RETOUR > 0) {Envoyer_Consigne(RETOUR); RETOUR = 0;} // Configuration de retour des fonctions à risque.
  //------------------------------------------------------------------------------- 
  if (Mode_en_cours == 1) { // Menu EXPLOITATION. : Signale la fin du gradateur par le code 53.
      Envoyer_Consigne(53);
      if (!Mode_MANUEL) display.update(); // Efface la visualisation des valeurs ligne du bas.
      // Codes 16 et 17 : Affichage de la posture actuelle ou de la posture SAV en EEPROM.
      if ((Num_Exploitation == 16) || (Num_Exploitation == 17)) Regenere_le_menu_posture();
      if  ((Num_Exploitation == 2) || (Num_Exploitation == 12) || (Num_Exploitation == 13))
          {Affiche_Ecran_MENU(); Affiche_Ecran_EXPLOITER(); display.update();}}
  //-------------------------------------------------------------------------------         
  if (Mode_en_cours == 3) { // Menu des OPTIONS.
      switch (Num_des_options) { // Change l'état du booléen système.
        case 1 : {Envoyer_Consigne(26); break;} // Disjoncteur énergie sur OFF.
        case 2 : {Envoyer_Consigne(31); break;} // Moteurs ON.
        case 3 : {Envoyer_Consigne(34); break;} // Mouvements lents.
        case 4 : {Envoyer_Consigne(36); Positionne_LED_Stab_Gyroscopique(); break;} // Fin de Stab. Gyroscopique.
        case 5 : {FIN_affichage_en_continu(); break;}} // Stoppe l'affichage continu.
        Affiche_etat_option_en_cours();}
  //-------------------------------------------------------------------------------
  if (Mode_en_cours == 6) { // Menu APPRENTISSAGE.
      if (Num_Apprentissage == 8) {AFF_Commande_pour_Apprentissage(); Envoyer_Consigne(72);
           Eteindre_Apprentissage(); BIT_Option = 7; Affiche_etat_option_en_cours();}
      if (Num_Apprentissage == 9) Change_le_num_PGM_cible = false;
      Eteindre_SECURITE();}
  //-------------------------------------------------------------------------------
  digitalWrite(LED_Retour_par_FIN,LOW);}
  
//----------------- Routines pour l'option de rérérence Gyroscopique ou Magnétique -----------------
void Traite_saisie_cap_magnerique_de_reference() {
  if (Sens_positif) {
      if (Poids_decimal == 1) Cap_magnetique_de_consigne++;
      if (Poids_decimal == 10) Cap_magnetique_de_consigne = Cap_magnetique_de_consigne + 10;
      if (Poids_decimal == 100) Cap_magnetique_de_consigne = Cap_magnetique_de_consigne + 100;
      if (Cap_magnetique_de_consigne > 359) Cap_magnetique_de_consigne = 0;}
  else {if (Poids_decimal == 1) Cap_magnetique_de_consigne--;
        if (Poids_decimal == 10) Cap_magnetique_de_consigne = Cap_magnetique_de_consigne - 10;
        if (Poids_decimal == 100) Cap_magnetique_de_consigne = Cap_magnetique_de_consigne - 100;
        if (Cap_magnetique_de_consigne < 0) Cap_magnetique_de_consigne = 0;} 
  AFF_le_CAP_de_reference();}
  
void AFF_le_CAP_de_reference() {
  Effacer_zone(78,37,3); display.setCursor(78,37);
  if (Cap_magnetique_de_consigne < 10 ) Affiche_0(); if (Cap_magnetique_de_consigne < 100 ) Affiche_0();
  display.print(Cap_magnetique_de_consigne); display.update();}

void Saisie_du_CAP_magnetique() {
  LED_Retour_par_FIN_clignote = true;
  Demande_Information(96); Cap_magnetique_de_consigne = Chaine_Memorisee.toInt();
  Prepare_afficher_item(); // Réalise effacer_zone(1, 37, 21); display.setCursor(1,37)
  Aff_TEXTE_EEPROM(863,12); AFF_le_CAP_de_reference(); Affiche_Degre(); Poids_decimal = 10;
  Mode_saisie_de_CAP_magnetique = true; Affiche_le_curseur();}
  
void Efface_le_curseur() {Effacer_zone(78,46,3);}
  
void Affiche_le_curseur() {
  X = 78; Efface_le_curseur(); if (Poids_decimal == 10) X = 84; if (Poids_decimal == 1) X = 90;
  display.setCursor(X,46); display.print(char(24)); display.update();}
  
void Fermer_la_saisie_du_CAP_magnetique() {
  Serial.print('C'); Serial.print(Cap_magnetique_de_consigne); Envoyer_etoile();
  Attendre_une_chaine_sur_le_Maitre(); Efface_le_curseur(); display.update();
  Mode_saisie_de_CAP_magnetique = false; LED_Retour_par_FIN_clignote = false;}
  
//------------------ Routines pour l'affichage graphique des données de navigation -----------------
void Ouvre_AFF_en_continu() {
  Affichage_NAV_Continu = true; Ecran_AFF_en_continu(); Serial.print(91); Envoyer_etoile();
  Eteindre_SECURITE();} // Active l'affichage en continu des données de navigation.
                    
void Limite_Ruban() {if (Ruban < -10) Ruban = -10; if (Ruban > 10) Ruban = 10;}

void Prepare_zone_aff_numerique(byte X) {Effacer_zone(86,X,4); display.setCursor(86,X);}

void Affiche_ecran_graphique() {
  //---------------------------- Affiche la rose des CAPs ----------------------
  display.drawCircle(32, 40, 23, WHITE); // Affiche le cercle.
  //--------- Tracer les graduations tous les 30 degrés ---------
  for (int Angle = 0; Angle < 360; Angle = Angle + 30) {
       Alpha = radians(Angle);
       X = 32 + (16 * sin(Alpha)); Y = 40 - (16 * cos(Alpha));
       SAVX = 32 + (23 * sin(Alpha)); SAVY = 40 - (23 * cos(Alpha));
       display.drawLine(X, Y, SAVX, SAVY, WHITE);}
  display.drawRect(16, 0, 33, 16, WHITE); // Cadre pour l'affichage numérique du CAP.
  //------------------------------ Ruban pour le Roulis ------------------------
  display.drawRect(55, 5, 73, 6, WHITE);
  for (byte X=55; X<128; X=X+18) display.drawLine(X, 0, X, 5, WHITE);
  //----------------------------- Ruban pour le Tangage ------------------------
  display.drawRect(121, 19, 7, 45, WHITE);
  for (byte Y=19; Y<64; Y=Y+11) display.drawLine(115, Y, 120, Y, WHITE);
  //------------------- Ecriture des divers champs de données ------------------
  display.setCursor(62,18); Aff_TEXTE_EEPROM(439,3); display.setCursor(62,31);
  Aff_TEXTE_EEPROM(992,3); display.setCursor(62,43);
  if (Reference_gyroscopique) {Aff_TEXTE_EEPROM(944,1); Aff_TEXTE_EEPROM(996,3);}
  else Aff_TEXTE_EEPROM(995,4);
  //--------------- Affichage de la valeur de la référence C ou G --------------
  if (Reference_gyroscopique) { 
  Demande_Information(57); Calage_Gyroscopique = Chaine_Memorisee.toInt(); display.print(Calage_Gyroscopique);} 
  else {Demande_Information(96); Cap_magnetique_de_consigne = Chaine_Memorisee.toInt();
        display.print(Cap_magnetique_de_consigne);} Affiche_Degre();
  display.setCursor(56,55); Aff_TEXTE_EEPROM(999,3); // Affiche "E = "  
  //------------------ Tracer le "bug" du calage gyroscopique ------------------
  if (Reference_gyroscopique) Alpha = radians(Calage_Gyroscopique);
  else Alpha = radians(Cap_magnetique_de_consigne);
  X = 32 + (19 * sin(Alpha)); Y = 40 - (19 * cos(Alpha));
  display.drawCircle(X, Y, 2, WHITE);
  //----------------------------------------------------------------------------
  X = 33; Y = 41;} // Pour ne pas effacer des pixel au premier affichage du rayon.
  
void Rafraichir_les_donnees_de_navigation() {
  if (Mode_Graphique) {
      Attendre_une_chaine_sur_le_Maitre(); Angle = Chaine_Memorisee.toInt();
      //------------------------- Calcule les coordonnées ---------------------- 
      Alpha = radians(Angle);
     SAVX = X; SAVY = Y; X = 32 + (15 * sin(Alpha)); Y = 40 - (15 * cos(Alpha));
     //--------------------- Trace le rayon vecteur du CAP --------------------
     display.drawLine(32, 40, SAVX, SAVY, BLACK);
     display.drawLine(32, 40, X, Y, WHITE);
     //------------------ Affiche la valeur numérique du CAP ------------------
     Position_aff_CAP = 22;
     if (Angle < 100) Position_aff_CAP = 25; if (Angle < 10) Position_aff_CAP = 29;
     Effacer_zone(22, 4, 4); display.setCursor(Position_aff_CAP,4);
     display.print(Angle); Affiche_Degre(); Envoyer_etoile();
     //------------------- Affiche le Tangage en mode ruban -------------------
     Attendre_une_chaine_sur_le_Maitre(); Tangage = Chaine_Memorisee.toInt();
     for (byte X=122; X<127; X++) display.drawLine(X, 20, X, 62, BLACK); // Efface le ruban.
     if (Tangage > 90) Tangage = -90; Ruban = Tangage; Limite_Ruban(); Ruban = map(Ruban,-10,+10,63,19);
     for (byte X=122; X<127; X++) display.drawLine(X, 41, X, Ruban, WHITE); // Trace le ruban.
     //-------------------- Affiche le Tangage en numérique -------------------
     Prepare_zone_aff_numerique(31); if (Tangage > 0) display.print('+');
     display.print(Tangage); Affiche_Degre(); Envoyer_etoile();
     //-------------------- Affiche le Roulis en mode ruban -------------------
     Attendre_une_chaine_sur_le_Maitre(); Roulis = Chaine_Memorisee.toInt();
     for (byte Y=6; Y<10; Y++) display.drawLine(56, Y, 126, Y, BLACK); // Efface le ruban.
     if (Roulis > 90) Roulis = -90; Ruban = Roulis; Limite_Ruban(); Ruban = map(Ruban,-10,+10,55,127);
     for (byte Y=6; Y<10; Y++) display.drawLine(91, Y, Ruban, Y, WHITE); // Trace le ruban.
     //-------------------- Affiche le Roulis en numérique -------------------
     Prepare_zone_aff_numerique(18); if (Roulis > 0) display.print('+');
     display.print(Roulis); Affiche_Degre(); Envoyer_etoile();
     //--------------- Passer la valeur du calage gyroscopique --------------- 
     Attendre_une_chaine_sur_le_Maitre(); Envoyer_etoile();
     //----------- Affiche l'écart de route / à la référence G ou M ----------
     Attendre_une_chaine_sur_le_Maitre(); Ecart_gyroscopique = Chaine_Memorisee.toInt();
     Effacer_zone(80,55,5); display.setCursor(80,55);
     if(!Reference_gyroscopique) Recalculer_Ecart_de_route();
     if (Ecart_gyroscopique > 0) display.print('+'); display.print(Ecart_gyroscopique);      
     Affiche_Degre(); Envoyer_etoile();}
  else {Attendre_une_chaine_sur_le_Maitre(); Affiche_le_CAP(); Envoyer_etoile(); 
        Attendre_une_chaine_sur_le_Maitre(); Affiche_le_Tangage(); Envoyer_etoile();
        Attendre_une_chaine_sur_le_Maitre(); Affiche_le_Roulis(); Envoyer_etoile();
        Attendre_une_chaine_sur_le_Maitre(); Affiche_le_Calage_GYRO(); Envoyer_etoile();
        Attendre_une_chaine_sur_le_Maitre(); Affiche_Ecart_route() ; Envoyer_etoile();}
  display.update();}
  
void Recalculer_Ecart_de_route() {// Angle contient la valeur du cap magnétique actuel.
  // Carculer l'amplitude de l'écart de route.
  Ecart_gyroscopique = 0; // Contiendra la valeur de l'écart de route magnétique.
  while (Angle != Cap_magnetique_de_consigne) {
    Ecart_gyroscopique++; Angle++; if (Angle > 359) Angle = 0;}
    if (Ecart_gyroscopique > 180) Ecart_gyroscopique = (360 - Ecart_gyroscopique) * (-1);}

//-------------------- Routines de servitude pour traiter le capteur incrémental -------------------
void Delay_antirebonds() {for(int I=0; I<100; I++);} // Valeur de 3 expérimentale.

void Front_sur_A(){ // Interruption quand A change d'état avant B.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds.
  if (digitalRead(Codeur_A) != A_set ) {// Test pour savoir si l'état a changé.
    A_set = !A_set;
    if (A_set && !B_set) {Sens_positif = true; Rotation = true; Delay_antirebonds();} } }

void Front_sur_B(){ // Interruption quand B change d'état avant A.
  if (INT_active) Delay_antirebonds(); // Attendre un peu la fin des rebonds. 
  if( digitalRead(Codeur_B) != B_set ) {// Test pour savoir si l'état a changé.
    B_set = !B_set;
    // Décrémenter le compteur si B avant A.
    if (B_set && !A_set) {Sens_positif = false; Rotation = true; Delay_antirebonds();} } }
//--------------------------------------------------------------------------------------------------
  
void AFF_TITRE_Apprentissage() {Curseur_32_4(); Aff_TEXTE_EEPROM(187,7); Aff_TEXTE_EEPROM(954,3);}

void MENU_APPRENTISSAGE() {
  Mode_en_cours = 6;
  NB_a_incrementer_decrementer = Num_Apprentissage; Incrementer_Decrementer(9);
  Num_Apprentissage = NB_a_incrementer_decrementer;
  AFF_la_commande_du_menu_apprentissage();}
  
void AFF_la_commande_du_menu_apprentissage() {
  TITRE_MENU_et_AFF_TITRE_APPRENTsge();
  Curseur_0_30();  
  switch (Num_Apprentissage) {
    case 1 : {Aff_TEXTE_EEPROM(809,7); Aff_TEXTE_EEPROM(802,8); Affiche_Num_PGM_cible(); break;} // Liste le PGM.
    case 2 : {Aff_TEXTE_EEPROM(816,21); break;} // Effacer le dernier code enregistré.
    case 3 : {Aff_TEXTE_EEPROM(795,15); Affiche_Num_PGM_cible(); break;} // Effacer le PGM indexé en EEPROM.
               // Ci-dessous : Le PGM actuel devient la SOURCE d'un transfert.    
    case 4 : {Aff_TEXTE_EEPROM(806,4); Demande_Information(84); display.print(Chaine_Memorisee);
              Aff_TEXTE_EEPROM(714,3); AFF_le_mot_SOURCE(); ESP_Point_Int(); break;}    
    case 5 : {Aff_TEXTE_EEPROM(858,5); AFF_le_mot_SOURCE(); Aff_TEXTE_EEPROM(5,6);
              Affiche_Num_PGM_cible();break;} // Transférer la SOURCE dans le PGM actuellement indexé.
    case 6 : {Aff_TEXTE_EEPROM(837,7); Aff_TEXTE_EEPROM(802,8); Affiche_Num_PGM_cible(); break;} // Déclencher PGM.
    case 7 : {Aff_TEXTE_EEPROM(977,15); AFFICHE_RETOUR(); break;} // Imposer le Nb de chainages de PGM.
    case 8 : {AFF_Commande_pour_Apprentissage(); break;}  // Mode apprentissage sur la SONDE O/N.
    case 9 : {AFF_TXT_EEPROM_et_RETOUR(43,19); break;}} // Change de PGM cible en EEPROM.
  display.update();}
  
void Positionne_LED_Stab_Gyroscopique() {
  Demande_Information(44); // Retourne l'état des booléens système.
  if (Chaine_Memorisee.charAt(3) == '1') digitalWrite(LED_Stabilisation_Gyroscopique,HIGH);
  else digitalWrite(LED_Stabilisation_Gyroscopique,LOW);}
  
void FIN_affichage_en_continu() {
  while (Serial.available()) Caractere = Serial.read(); // Vider le FIFO.
  Affichage_NAV_Continu = false; Envoyer_Consigne(92); display.clear(); Affiche_Ecran_MENU();
  Num_des_options = 4; Sens_positif = true; MENU_des_OPTIONS(); LED_Retour_par_FIN_clignote = false;}
  
void Ecran_AFF_en_continu() {
  LED_Retour_par_FIN_clignote = true; display.clear();
  if (Mode_Graphique) {Affiche_ecran_graphique();}
  else {CADRE(); display.drawLine(1, 28, 126, 28, WHITE); Affiche_cadre_du_haut(10);
        Aff_TEXTE_EEPROM(936,18); Affiche_ecran_pour_la_navigation();}}

//-------------------------- Gestion de l'apprentissage et de son l'utilisation --------------------
void AFF_Commande_pour_Apprentissage() {
  Curseur_0_30(); Aff_TEXTE_EEPROM(957,20); BIT_Option = 7; Affiche_etat_option_en_cours();}
  
void Realiser_un_PGM() {
  Effacer_ACR(); // Pour contrôler le "OK" de la fin d'exécution du programme.
  //--------------------- Déclencher la réalisation du PGM ------------------------
  Demande_Information(87); // Déclenche le programme sans vérifier un retour "OK".
  if (Chaine_Memorisee != "E18") {
  //--------------- Barrer les codes dans la grille de listage ----------------
  byte C=1; byte L=26;
  while ((Chaine_Memorisee != "OK") && (Chaine_Memorisee != "E20")) {
    display.drawLine(C, L, C+16, L-11, WHITE); display.update(); // barre la case et raffraichi l'écran.
    L = L+12; if (L > 62) {L = 26; C = C+16;} // Change de case.
    Attendre_une_chaine_sur_le_Maitre();}} // Attente du code suivant.
    AFF_ACR();}
  
void ERR_NON_Securite() {BIP_et_XY(); display.print("E12");}

void BIP_et_XY() {BIP(); display.setCursor(4,4);}

boolean Apprentissage_en_cour() {
  Demande_Information(44); // Retourne l'état des booléens système.
  return Chaine_Memorisee.charAt(7) == '1';}

void Transferer_la_SOURCE() {
  if (!Apprentissage_en_cour()) {
      if (digitalRead(LED_SECURITE)) {Envoyer_Consigne(13); Eteindre_SECURITE();}
      else ERR_NON_Securite();}
  else {BIP_et_XY(); display.print("E22");}}
  
void Effacer_un_PGM() {
  if (!Apprentissage_en_cour()) {
      if (digitalRead(LED_SECURITE)) {Envoyer_Consigne(88); Eteindre_SECURITE();}
      else ERR_NON_Securite();}
  else {BIP_et_XY(); display.print("E21");}}
  
void Effacer_le_dernier_code_appris() {
  if (Apprentissage_en_cour()) Envoyer_Consigne(89);
  else {Eteindre_SECURITE(); BIP_et_XY(); display.print("E16");}}
  
void Affiche_Num_PGM() {
  display.drawRect(114, 41, 13, 9, WHITE);
  display.drawRect(113, 40, 14, 11, WHITE);
  display.setCursor(115,42); display.setTextColor(BLACK, WHITE); display.print('P');
  display.print(Chaine_Memorisee); display.setTextColor(WHITE, BLACK);}

void Affiche_num_du_PGM_cible() {Demande_Information(84); Affiche_Num_PGM();}
  
void Lister_un_PGM() {
  byte C=3; byte L=18; Effacer_zone(1, 29, 21); // Efface la ligne centrale de texte.
  Prepare_afficher_item(); // Efface la ligne centrale de texte.
  //----------------------- Affiche la grille de listage --------------------------
  CADRE(); for (byte C=16; C<113; C=C+16) display.drawLine(C, 17, C, 62, WHITE);
  for (byte L=15; L<63; L=L+12) display.drawLine(1, L, 127, L, WHITE);
  //------------------------- Affiche le n° du PGM ciblé --------------------------
  Affiche_num_du_PGM_cible();
  //---------------------- Réceptionne et liste les codes -------------------------
  Serial.print("86*"); AFF_le_code_en_bas_a_droite(86);
  while (!Fin_de_listage()) {
         Attendre_une_chaine_sur_le_Maitre(); display.setCursor(C,L);
         if (!Fin_de_listage()) 
             {if (Chaine_Memorisee.length() < 2) Affiche_0();
                  display.print(Chaine_Memorisee); display.update();}
         L = L+12; if (L > 62) {L = 18; C = C+16;}}
   if (Chaine_Memorisee == "E15") {BIP_et_XY(); display.print("E15");}}
  
boolean Fin_de_listage() {return ((ACR_egal_OK()) || (Chaine_Memorisee == "E15"));} 

void Affiche_Num_PGM_cible() {
  Demande_Information(84); display.print(Chaine_Memorisee); ESP_Point_Int();}
  
void AFF_num_PGM_actif() {
  if (Change_le_chainage_PGM) {Chaine_Memorisee = "N:" + String(NB_PGM_a_chainer);
      EFF_ACR_et_Positionne_XY(); display.print(Chaine_Memorisee); display.update();}
  else {Demande_Information(84); // demande le n° de PGM actif.
        Num_PGM = Chaine_Memorisee.toInt(); Chaine_Memorisee = "P:" + Chaine_Memorisee; AFF_ACR();}}

void Prepare_changer_PGM_ou_Chainage() {
  Gradateur = true; AFF_num_PGM_actif(); LED_Retour_par_FIN_clignote = true;}
  
void Changer_de_PGM_EEPROM() {Change_le_num_PGM_cible = true; Prepare_changer_PGM_ou_Chainage();}
  
void Chainer_des_PGM() {Change_le_chainage_PGM = true; Prepare_changer_PGM_ou_Chainage();}

void Imposer_le_num_du_PGM_cible() {
  Serial.print("85P"); Serial.print(Num_PGM); Envoyer_etoile(); Attendre_une_chaine_sur_le_Maitre();}
  
void Enchainer_les_PGMs() {
  //-------------------- Vérifier que le premier programme est NON vide --------------------
  Demande_Information(94); // ACR retourné = "OK" si NON vide.
  if (ACR_egal_OK()) {
      SAV_Num_PGM = Num_PGM;
      while (NB_PGM_a_chainer > 0) {
        display.clear(); CADRE_UP();
        Lister_un_PGM(); Realiser_un_PGM(); Num_PGM++; // Passe au programme suivant.
        //---------- Sortir de la boucle si Pb en exécution de PGM ou débordement de n°9 ----------
        if((Chaine_Memorisee != "OK") || (Num_PGM > 9)) NB_PGM_a_chainer = 0;
        //---------------------------- Si "OK" passer au PGM suivant ------------------------------
        else {NB_PGM_a_chainer--; if (NB_PGM_a_chainer > 0) Imposer_le_num_du_PGM_cible();}
        //-----------------------------------------------------------------------------------------
        if (NB_PGM_a_chainer > 0) {
            Demande_Information(94); //Vérifie que NON vide.
            if (Chaine_Memorisee != "OK") {NB_PGM_a_chainer = 0; AFF_ACR();
                Chaine_Memorisee = String(Num_PGM); Affiche_Num_PGM(); display.update(); ATTENDRE_un_BP();}}}
      Num_PGM = SAV_Num_PGM; Imposer_le_num_du_PGM_cible(); // Restituer le n° du PGM indexé en EEPROM.
      NB_PGM_a_chainer = 1; Curseur_32_4(); Aff_TEXTE_EEPROM(599,4); Aff_TEXTE_EEPROM(984,9);}
  //----------------------------------------------------------------------------------------
  else AFF_ACR();}

//--------------------------------------------------------------------------------------------------
void Affiche_moteur_actif() {
  Effacer_zone(55,23,1); display.setCursor(7,23); Aff_TEXTE_EEPROM(717,8); display.print(Jambe);
  Prepare_afficher_item(); // Réalise effacer_zone(1, 37, 21); display.setCursor(1,37)
  Curseur_1_37(); Aff_TEXTE_EEPROM(708,9);
  switch ( Articulation) { 
    case 1 : {Aff_TEXTE_EEPROM(725,6); break;}
    case 2 : {Aff_TEXTE_EEPROM(731,5); break;}
    case 3 : Aff_TEXTE_EEPROM(736,6);}
  display.update();}
  
void Inverse_LED_SECURITE() {digitalWrite(LED_SECURITE,!digitalRead(LED_SECURITE));}
  
void Eteindre_SECURITE() {digitalWrite(LED_SECURITE,LOW);}

void Eteindre_Apprentissage() {digitalWrite(LED_Apprentissage,LOW);}
              
void Purge_les_booleens() {
  LED_Retour_par_FIN_clignote = false; LED_Erreur_clignote = false; Mode_pointage_LASER = false;
  Gradateur = false; Mode_MANUEL = false; Change_le_chainage_PGM = false;}
    
void Teste_Touche_PHARES (){
  if (BP == 13) {PHARES = !PHARES; if (PHARES) Envoyer_Consigne(27); else Envoyer_Consigne(28);}}  
 
//-------------------------------- Récupération des données de POSTURE -----------------------------
void AFF_POSTURE(byte Code_consigne) {
  Prepare_afficher_item(); Demande_Information(Code_consigne);
  Recupere_et_affiche_la_posture(8,20); Recupere_et_affiche_la_posture(8,37); 
  Recupere_et_affiche_la_posture(8,55); Recupere_et_affiche_la_posture(41,20);
  Recupere_et_affiche_la_posture(41,37); Recupere_et_affiche_la_posture(41,55);
  Recupere_et_affiche_la_posture(72,20); Recupere_et_affiche_la_posture(72,37);
  Recupere_et_affiche_la_posture(72,55); Recupere_et_affiche_la_posture(104,20);
  Recupere_et_affiche_la_posture(104,37); Efface_le_code();
  display.setCursor(104,55); display.print(Chaine_Memorisee);
  LED_Retour_par_FIN_clignote = true;}
  
void Recupere_et_affiche_la_posture (byte X, byte Y) {
  display.setCursor(X,Y); display.print(Chaine_Memorisee);
  Attendre_une_chaine_sur_le_Maitre();}
  
void Regenere_le_menu_posture() {
  Prepare_afficher_item(); Effacer_zone(8,20,20); Effacer_zone(8,55,20); Efface_le_code();
  if (Num_Exploitation == 16) Aff_TEXTE_EEPROM(742,18); else Affiche_item_posture_EEPROM(); display.update();}
  
void AFF_graduations_telemetrie(byte Decalage_vertical) {
  for (byte I=12; I<120; I=I+24) display.drawLine(I, 17 + Decalage_vertical, I, 21 + Decalage_vertical, WHITE);
  if (Decalage_vertical > 0) Decalage_vertical = 29; display.setCursor(54,21 + Decalage_vertical);
  display.print("1m"); display.setCursor(102,21 + Decalage_vertical); display.print("2m");}
  
//-------------------------- Gestion de l'affichage du balayage télémétrique ----------------------- 
void Affiche_le_balayage_ultrasons() {
  Prepare_afficher_item(); CADRE(); LED_Retour_par_FIN_clignote = true;
  //--------------------- Trace les graduations de distance ---------------------
  AFF_graduations_telemetrie(0); AFF_graduations_telemetrie(41);
  //-----------------------------------------------------------------------------
  display.update(); Demande_Information(70);
  RETOUR = 24; // Position verticale du tracé. (On réutilise cette variable disponible durant cette procédure.)
  for (byte I = 0; I<31; I++) {// Lecture d'une donnée et tracé de sa ligne représentative.
       Affiche_une_ligne_representative(); Attendre_une_chaine_sur_le_Maitre();}
  Affiche_une_ligne_representative(); display.update(); RETOUR = 0;} // Pour désarmer l'action sur RETOUR > 0.
 
 void Affiche_une_ligne_representative() {
   CNA = Chaine_Memorisee.toInt(); CNA = CNA * 2; // Pour une portée de 2m sur la largeur d'affichage.
   for (CNA; CNA > 0; CNA--) display.drawPixel(CNA+11, RETOUR, WHITE); RETOUR++;}

//-------------------------------- Gestion du spectre colorimétrique -------------------------------  
void AFF_MENU_Spectre_colorimetrique() {
  Prepare_afficher_item(); Demande_Information(68); LED_Retour_par_FIN_clignote = true;
  Recupere_et_affiche_couleur(5,18,635,17); // Lumiere totale.
  Recupere_et_affiche_couleur(5,30,309,6); // Bleu.
  Recupere_et_affiche_couleur(5,42,315,6); // Mauve.
  Recupere_et_affiche_couleur(5,54,321,6); // Rouge.
  Recupere_et_affiche_couleur(71,30,327,6); // Orange.
  Recupere_et_affiche_couleur(71,42,622,6); // Jaune. 
  Efface_le_code(); display.setCursor(71,54); Aff_TEXTE_EEPROM(628,6); display.print(Chaine_Memorisee);} // Vert.
  
void Recupere_et_affiche_couleur(byte X, byte Y, int PTR, byte Lgr) {
  display.setCursor(X,Y); Aff_TEXTE_EEPROM(PTR,Lgr); display.print(Chaine_Memorisee);
  Attendre_une_chaine_sur_le_Maitre();}

//---------------------------------------- Gestion des MENUs --------------------------------------- 
void MENU_POSTURES(){
  NB_a_incrementer_decrementer = Num_des_postures; Incrementer_Decrementer(9);
  Num_des_postures = NB_a_incrementer_decrementer;
  Affiche_Ecran_POSTURES(); display.update();}
  
void AFF_mot_JEKERT() {Aff_TEXTE_EEPROM(552,6);} 
  
void Affiche_Ecran_POSTURES(){
  display.setCursor(23,21); Aff_TEXTE_EEPROM(30,13); Effacer_zone(6, 55, 17);
  Prepare_afficher_item();
  switch (Num_des_postures) {
    case 1 : {AFF_mot_JEKERT(); Espace(); display.print(char(25));  // La sonde se pose sur le sol.
              Aff_TEXTE_EEPROM(285,7); break;}
    case 2 : {AFF_TXT_EEPROM_et_RETOUR(452,18); break;} // Stable Raisonnable.
    case 3 : {AFF_TXT_EEPROM_et_RETOUR(470,16); break;} // Hauteur Maximale.
    case 4 : {Aff_TEXTE_EEPROM(486,20); break;} // Moteurs au Neutre opérationnel.
    case 5 : {Aff_TEXTE_EEPROM(506,17); break;} // Restituer la posture apprise en EEPROM.
    case 6 : {BIP(); Aff_TEXTE_EEPROM(535,9); Aff_TEXTE_EEPROM(602,12); break;}  // Décollage.
    case 7 : {Aff_TEXTE_EEPROM(523,12); break;} // Atterrissage.
    case 8 : {Aff_TEXTE_EEPROM(544,14); break;} // Déposer la Sonde.
    case 9 : {Aff_TEXTE_EEPROM(558,18);}}       // Stable Transversal.
    if (Num_des_postures != 6) display.print('.');}
     
void MENU_EXPLOITER(){
  NB_a_incrementer_decrementer = Num_Exploitation; Incrementer_Decrementer(19);
  Num_Exploitation = NB_a_incrementer_decrementer;
  Affiche_Ecran_EXPLOITER(); display.update();}
 
void Affiche_Ecran_EXPLOITER(){
  Prepare_afficher_item();
  switch (Num_Exploitation) {
    case  1 : {Aff_TEXTE_EEPROM(333,3); Aff_TEXTE_EEPROM(661,14); break;} // Enregistrer un balayage télémètre.   
    case  2 : {Aff_TEXTE_EEPROM(354,3); Aff_TEXTE_EEPROM(661,14); break;} // Afficher le balayage télémétrique.
    case  3 : {Aff_TEXTE_EEPROM(251,19); break;} // Sauvegarder la posture actuelle.
    case  4 : {Aff_TEXTE_EEPROM(270,7); ESP_Point_Int(); break;} // QUITTER.
    case  5 : {Aff_TEXTE_EEPROM(393,10); AFF_mot_JEKERT(); break;} // Réveiller la sonde.    
    case  6 : {Aff_TEXTE_EEPROM(389,4); Aff_TEXTE_EEPROM(167,11); break;} // Test communication avec le MPU6050.
    case  7 : {Affiche_Calage_Gyroscope(); break;}  // Calage du gyroscope.    
    case  8 : {Aff_TEXTE_EEPROM(426,20); break;} // Phare et LASER = 127.    
    case  9 : {Aff_TEXTE_EEPROM(403,17); break;} // Gradateur Phare.
    case 10 : {Aff_TEXTE_EEPROM(403,10); Aff_TEXTE_EEPROM(420,6); break;} // Gradateur LASER.
    case 11 : {Aff_TEXTE_EEPROM(378,11); Aff_TEXTE_EEPROM(372,6); // Utiliser le LASER.
               ESP_Point_Int(); AFFICHE_RETOUR(); break;}
    case 12 : {Aff_TEXTE_EEPROM(333,21);break;}  // Enregistrer un spectre colorimétrique.
    case 13 : {Aff_TEXTE_EEPROM(354,3); Aff_TEXTE_EEPROM(336,18); break;} // Afficher le spectre couleur.
    case 14 : {AFF_TXT_EEPROM_et_RETOUR(675,16); break;} // TORSION active.
    case 15 : {AFF_TXT_EEPROM_et_RETOUR(691,17); break;} // Pilotage en MANUEL.
    case 16 : {Aff_TEXTE_EEPROM(742,18); break;} // Affiche "Posture actuelle ?".
    case 17 : {Affiche_item_posture_EEPROM(); break;} // Affiche "Posture en EEPROM ?".
    case 18 : {Aff_TEXTE_EEPROM(63,4); Aff_TEXTE_EEPROM(373,5); ESP_Point_Int(); break;} // Affiche "TIR LASER ?".
    case 19 : {Aff_TEXTE_EEPROM(863,4); Aff_TEXTE_EEPROM(1002,10); ESP_Point_Int();}}} // saisie REF. Magnétique.
    
void AFF_le_mot_SOURCE() {Aff_TEXTE_EEPROM(850,6);}

void Affiche_item_posture_EEPROM() {Aff_TEXTE_EEPROM(742,8); Aff_TEXTE_EEPROM(514,9); ESP_Point_Int();}

void AFF_TXT_EEPROM_et_RETOUR(int PTR, byte LGR) {Aff_TEXTE_EEPROM(PTR,LGR); AFFICHE_RETOUR();}
  
void MENU_des_DONNEES(){
  NB_a_incrementer_decrementer = Num_des_donnees; Incrementer_Decrementer(5);
  Num_des_donnees = NB_a_incrementer_decrementer;
  Effacer_zone(7, 19, 19); Effacer_zone(5, 32, 20); Effacer_zone(5, 42, 20); Effacer_zone(5, 52, 20);
  CADRE(); display.drawLine(1, 28, 126, 28, WHITE);
  Affiche_Ecran_des_DONNEES(); display.update();}
 
void Affiche_Ecran_des_DONNEES(){
  AFF_num_PGM_actif();
  switch (Num_des_donnees) {
              // Ci dessous : Page Phares et LASER.
    case 1 : {display.setCursor(15,19); Aff_TEXTE_EEPROM(760,16);
              display.setCursor(19,32); Aff_TEXTE_EEPROM(363,15);
              Demande_Information(44); // Retourne l'état des booléens système.
              display.setCursor(25,42); OUI_NON(5); display.setCursor(85,42); OUI_NON(6);
              Demande_Information(46); display.setCursor(25,52); display.print(Chaine_Memorisee);
              Demande_Information(47); display.setCursor(85,52); display.print(Chaine_Memorisee); break;}
              // Version du PGM de la Sonde, État motorisation et état Hibernation.
    case 2 : {Demande_Information(24); // Version du PGM de la Sonde et état d'activité.
              display.setCursor(10,19); Aff_TEXTE_EEPROM(202,8); display.print(Chaine_Memorisee);             
              Demande_Information(50); Curseur_5_32(); AFF_mot_JEKERT(); Aff_TEXTE_EEPROM(651,11);
              display.print(Chaine_Memorisee); // État hibernation.              
              Demande_Information(44); // Retourne l'état des booléens système.
              display.setCursor(5,42); Aff_TEXTE_EEPROM(220,17); OUI_NON(1);        
              display.setCursor(23,52); Aff_TEXTE_EEPROM(237,14); OUI_NON(2); break;}
              // Ci dessous : Page  Télémètre, Météo et Eclairages O/N.
    case 3 : {Curseur_7_19(); Demande_Information(51); display.print(Chaine_Memorisee);
              display.setCursor(7,32); Demande_Information(48); // Retourne la valeur de la Température.              
              display.print(Chaine_Memorisee); Affiche_Degre(); display.print('C');
              Demande_Information(49); // Retourne la valeur de l'hygrométrie relative.
              Curseur_7_42(); display.print(Chaine_Memorisee);
              Demande_Information(44); // Retourne l'état des booléens système.
              Curseur_7_52(); Affiche_un_Soleil(); Aff_TEXTE_EEPROM(209,11); OUI_NON(0); break;}               
               // Ci-dessous : U moteurs, Stabilisation / Bouclier / Apprentissage.
    case 4 : {Demande_Information(45);
              Curseur_7_19(); display.print(Chaine_Memorisee); Demande_Information(44);
              display.setCursor(7,32); Aff_TEXTE_EEPROM(162,16); OUI_NON(3);                                          
              Curseur_7_42(); Aff_TEXTE_EEPROM(277,16); OUI_NON(4);
              Curseur_7_52(); Aff_TEXTE_EEPROM(293,16); OUI_NON(7); break;}
              // Ci dessous : Les données de navigation, CAP, Tangage, Roulis, Ecart de route.
    case 5 : {Affiche_ecran_pour_la_navigation(); Affiche_les_donnees_de_navigation(); break;}}}
    
void AFF_chaine_et_degre() {display.print(Chaine_Memorisee); Affiche_Degre();}
    
void Affiche_les_donnees_de_navigation() {
  Demande_Information(52); Affiche_le_CAP(); Demande_Information(56); Affiche_le_Tangage();
  Demande_Information(93); Affiche_le_Roulis(); Demande_Information(57); Affiche_le_Calage_GYRO();
  Demande_Information(60); Affiche_Ecart_route() ;}
  
void Affiche_le_CAP() {Effacer_zone(89,18,4); display.setCursor(89,19); AFF_chaine_et_degre();}
void Affiche_le_Tangage() {Effacer_zone(42,32,3); display.setCursor(42,32); display.print(Chaine_Memorisee);}
void Affiche_le_Roulis() {Effacer_zone(107,32,3); display.setCursor(107,32); display.print(Chaine_Memorisee);}
void Affiche_le_Calage_GYRO() {Effacer_zone(100,42,4); display.setCursor(100,42); AFF_chaine_et_degre();}
void Affiche_Ecart_route() {Effacer_zone(94,52,5); display.setCursor(94,52); AFF_chaine_et_degre();}
  
void Affiche_ecran_pour_la_navigation() {
  display.setCursor(11,19); Aff_TEXTE_EEPROM(863,12); // Affiche "CAP actuel =".
  Curseur_5_32(); Aff_TEXTE_EEPROM(875,16); display.setCursor(11,42); // Affiche "Tang.      Roulis".
  Aff_TEXTE_EEPROM(892,14); display.setCursor(11,52); // Calage GYRO.
  Aff_TEXTE_EEPROM(906,13);} // Écart de route.
              
void Curseur_7_19() {display.setCursor(7,19);}
void Curseur_5_32() {display.setCursor(5,32);}
void Curseur_7_42() {display.setCursor(7,42);}
void Curseur_7_52() {display.setCursor(7,52);}
void Curseur_0_30() {display.setCursor(0,30);}
void Curseur_1_37() {display.setCursor(1,37);}
void Curseur_32_4() {display.setCursor(32,4);}

void MENU_des_OPTIONS() {
  NB_a_incrementer_decrementer = Num_des_options; Incrementer_Decrementer(5);
  Effacer_zone(50,48,5); // Efface le texte (OUI) ou (NON) si option Changer le PGM indexé en EEPROM.
  Num_des_options = NB_a_incrementer_decrementer; Effacer_zone(5, 30, 20); display.setCursor(5,30);
  switch (Num_des_options) {
    case 1 : {Aff_TEXTE_EEPROM(115,14); display.setCursor(72,30);
              Affiche_un_Soleil(); break;}  // Mode "couper les éclairages".
    case 2 : {Aff_TEXTE_EEPROM(129,13); break;}  // Moteurs OFF.
    case 3 : {Aff_TEXTE_EEPROM(142,20); break;}  // Mouvements rapides.
    case 4 : {Aff_TEXTE_EEPROM(162,17); break;}  // Stabilisation Gyroscopique.
    case 5 : {Aff_TEXTE_EEPROM(919,17); ESP_Point_Int(); break;}} // Affichage continu O/N.
  BIT_Option = Num_des_options; BIT_Option--; if (BIT_Option == 4) BIT_Option = 7;
  Affiche_etat_option_en_cours();}
  
void Prepare_TITRE_MENU() {
  Gradateur = false; digitalWrite(LED_Retour_par_FIN,LOW); digitalWrite(LED_MOUVEMENTS,LOW);
  display.clear(); CADRE_UP(); display.drawLine(100, 0, 100, 15, WHITE);
  display.drawLine(24, 0, 24, 15, WHITE); Afficher_Rot_ou_CLV();}
  
void TITRE_MENU_et_AFF_TITRE_APPRENTsge() {Prepare_TITRE_MENU(); AFF_TITRE_Apprentissage();}
  
void Affiche_Ecran_MENU() {
  Prepare_TITRE_MENU(); display.setCursor(34,4);
  switch (Mode_en_cours) {
    case 1 : {Aff_TEXTE_EEPROM(82,9); break;} // Affiche EXPLOITER dans le cadre jaune.
    case 2 : {Aff_TEXTE_EEPROM(21,9); break;} // Affiche POSTURES dans le cadre jaune.
    case 3 : {Aff_TEXTE_EEPROM(66,8); break;} // Affiche OPTIONS dans le cadre jaune.
    case 4 : {Aff_TEXTE_EEPROM(74,8); break;} // Affiche DONNEES dans le cadre jaune.
    case 5 : {Aff_TEXTE_EEPROM(11,10); digitalWrite(LED_MOUVEMENTS,HIGH); Afficher_le_NB_fois(); break;}}
  display.update();}           // Ci-dessus affiche MOUVOIR dans le cadre jaune.
     
void MENU_MOUVOIR(){
   display.setCursor(30,20); Aff_TEXTE_EEPROM(576,11); Prepare_afficher_item();
   Aff_TEXTE_EEPROM(97,18); display.update();}
  
//---------------------------------- Diverses procédures de service --------------------------------    
void Prepare_afficher_item() {Effacer_zone(1, 37, 21); Curseur_1_37();}

void OUI_NON(byte Position) {
  if (Chaine_Memorisee.charAt(Position) == '1') Aff_TEXTE_EEPROM(357,3); else Aff_TEXTE_EEPROM(360,3);}

void Affiche_etat_option_en_cours() {
  Effacer_zone(56, 48, 3); Demande_Information(44); // Etat des options.
  display.setCursor(50,48); display.print('('); OUI_NON(BIT_Option); display.print(')'); display.update();}
  
void AFFICHE_RETOUR() {
  display.setCursor(6,55); display.print(char(16)); /* Flèche vers la droite. */ Aff_TEXTE_EEPROM(587,15);
  if (Mode_en_cours != 2) display.print('.'); BIP();}
     
void Traite_Gradateur() {
  if ((Change_le_num_PGM_cible) || (Change_le_chainage_PGM)) {
     if (Change_le_num_PGM_cible) {
         if (Sens_positif) {Num_PGM++; if (Num_PGM == 10) Num_PGM = 1;}
         else {if (Num_PGM > 1) Num_PGM--; else Num_PGM = 9;}
         Imposer_le_num_du_PGM_cible();}
     else {if (Sens_positif) {NB_PGM_a_chainer++; if (NB_PGM_a_chainer == 10) NB_PGM_a_chainer = 1;}
           else {if (NB_PGM_a_chainer > 1) NB_PGM_a_chainer--; else NB_PGM_a_chainer = 9;}}
     AFF_num_PGM_actif();}
  else {if (Sens_positif) Envoyer_Consigne(40); else Envoyer_Consigne(41);
        if (((Num_Exploitation == 9) || Num_Exploitation == (10)) && (Mode_en_cours == 1))
              AFFICHE_puissance_eclairages();}}
      
void AFFICHE_puissance_eclairages() {
  if (Num_Exploitation == 9) Demande_Information(46); else Demande_Information(47); // Demande la puissance.
  Effacer_zone(50, 52, 3); display.setCursor(50,52); display.print(Chaine_Memorisee);
  for(byte X=1; X < 129; X = X + 9) display.drawLine(X, 17, X, 19, WHITE);
  Effacer_zone(1, 21, 21);  display.drawRect(0, 20, 128, 9, WHITE);
  for(byte Y=21; Y < 28; Y++) display.drawLine(1, Y, Chaine_Memorisee.toInt()/2, Y, WHITE); display.update();}

void Affiche_Calage_Gyroscope() {
  Effacer_zone(91,36,4); Demande_Information(57); Curseur_1_37();
  Aff_TEXTE_EEPROM(892,14); Espace(); display.print(Chaine_Memorisee); Affiche_Degre();}

void Afficher_Rot_ou_CLV() {
  Effacer_zone(106,4,3); display.setCursor(106,4);
  if (Mode_en_cours == 5) Aff_TEXTE_EEPROM(179,3); else Aff_TEXTE_EEPROM(614,3); display.update();}
  
void Incrementer_Decrementer(byte MAXI) {// La valeur minimale est toujours 1.
  if (Sens_positif) {NB_a_incrementer_decrementer++; 
                     if (NB_a_incrementer_decrementer > MAXI) NB_a_incrementer_decrementer = 1;}
  else { if (NB_a_incrementer_decrementer > 1) NB_a_incrementer_decrementer--;
         else NB_a_incrementer_decrementer = MAXI;}}
         
void Traite_Item_Menus() {
   if (!Change_le_num_PGM_cible)
       switch (Mode_en_cours) {
          case 1 : {MENU_EXPLOITER(); break;}
          case 2 : {MENU_POSTURES(); break;}
          case 3 : {MENU_des_OPTIONS(); break;}
          case 4 : {MENU_des_DONNEES(); break;}
          case 5 : {MENU_MOUVOIR(); break;}
          case 6 :  MENU_APPRENTISSAGE();}}
     
void Teste_Touche_LASER() {
     if (BP == 15) {LASER = !LASER; if (LASER) Envoyer_Consigne(29); else Envoyer_Consigne(30);}}
  
void Afficher_le_NB_fois() {
  Effacer_zone(66,55,1); display.setCursor(18,55); Aff_TEXTE_EEPROM(90,8);
  display.print(NB_Repetition_Mvt); Aff_TEXTE_EEPROM(617,5); display.update();}

//--------------------------------- Gestion du clavier à 16 touches --------------------------------
void Tester_le_clavier() {
   BP = 0; // Indique une non utilisation du clavier.
  Teste_Colonne(C1); Teste_Colonne(C2); Teste_Colonne(C3); Teste_Colonne(C4);}

void Teste_Colonne(byte Colonne) {
  digitalWrite(Colonne, HIGH); // Passe la colonne au +5Vcc.
  CNA = analogRead(Mesure_U_clavier);
  if (CNA > 90) {// Codage du BP sollicité.
      digitalWrite(LED_Clavier,HIGH); // Signale l'enfoncement d'une touche.
      BP = Colonne + 6; if (CNA > 286) BP = BP - 4; if (CNA > 511) BP = BP - 4; if (CNA > 718) BP = BP - 4;
      while (analogRead(Mesure_U_clavier) > 90) delay (5); delay (10); // Attendre le relâcher.
      digitalWrite(LED_Clavier,LOW);} // Signale le relâcher de la touche.
  digitalWrite(Colonne, LOW);} // Repasse la colonne à zéro volts.
  
void ATTENDRE_un_BP() {
    BP = 0; // Pour ne pas sortir prématurément.
    while (BP == 0) {
           Tester_le_clavier(); digitalWrite(LED_Clavier,!digitalRead(LED_Clavier)); delay(50);}
    digitalWrite(LED_Clavier,LOW);}

//--------------------------------------------------------------------------------------------------
void BIP() {
  pinMode(BUZZER, OUTPUT); tone(BUZZER,3000); digitalWrite(LED_Erreur,HIGH);
  delay(100); noTone(BUZZER); digitalWrite(LED_Erreur,LOW); pinMode(BUZZER, INPUT);}
  
void Envoyer_etoile() {Serial.print('*');}

void AFF_le_code_en_bas_a_droite(byte Code_consigne) {
  Efface_le_code(); display.setCursor(115,54); display.print(Code_consigne); display.update();}

void Attendre_une_chaine_sur_le_Maitre() {// Cette procédure attend une chaine sur la ligne USB.
  Chaine_Memorisee = ""; // Vider le tampon pour une nouvelle capture.
  Sentinelle_detectee = false;  // Sentinelle non reçue.
  Compteur_caracteres = 0;
  while (!Sentinelle_detectee) {
    if (Serial.available()) {Caractere = Serial.read(); // Puiser un caractère dans le FIFO.
    Sentinelle_detectee = (Caractere == '*'); // Si sentinelle détectée, informer la boucle de réception.
    // Puis concaténer ce caractère SAUF la sentinelle :
    if ((Compteur_caracteres < LGR_chaine) && (!Sentinelle_detectee)) 
       {Chaine_Memorisee = Chaine_Memorisee + Caractere; Compteur_caracteres++;}}}}    

void Demande_Information(byte Code_consigne) {
  Verifier_OK = false; Transmettre_le_code_consigne(Code_consigne);}

void Transmettre_le_code_consigne(byte Code_consigne) {  
  Serial.print(Code_consigne); Envoyer_etoile();
  //---------------- Affiche le Code en bas à droite sous conditions ----------------
  if ((Code_consigne != 44) && (BP != 4) && (BP != 0) && (Code_consigne != 84) && (!Mode_Graphique))
       if (!Mode_saisie_de_CAP_magnetique) AFF_le_code_en_bas_a_droite(Code_consigne);
  //--------------------------------------------------------------------------------
  Attendre_une_chaine_sur_le_Maitre();
  if (Verifier_OK) AFF_ACR();}
  
void EFF_ACR_et_Positionne_XY() {Effacer_ACR(); display.setCursor(4,4);}
      
void AFF_ACR() {
  if ((Chaine_Memorisee != "OK") && (Chaine_Memorisee.charAt(0) != 'P') &&
      (!Affichage_NAV_Continu)&& (Chaine_Memorisee.length() < 4)) BIP();
  EFF_ACR_et_Positionne_XY();  // Ci-dessus <4 pour ne pas afficher les ACR de type 01011010.
  if (Chaine_Memorisee.length() < 4) display.print(Chaine_Memorisee); else display.print("OK");
  display.update();}

void Effacer_ACR() {Effacer_zone(4, 4, 3); display.update();}
      
void Efface_le_code() {Effacer_zone(114,53,2);}
      
void Envoyer_Consigne(byte Code_consigne) {
  Verifier_OK = true; Transmettre_le_code_consigne(Code_consigne);}
  
boolean ACR_egal_OK() {return Chaine_Memorisee == "OK";}
    
void ESP_Point_Int() {Aff_TEXTE_EEPROM(200,2);}
void Affiche_Degre() {display.print(char(247));}
void Affiche_un_Soleil() {display.print(char(15));}
void Espace() {display.print(' ');}
void Affiche_0() {display.print('0');}

//--------------------------------- Routines pour les affichages -----------------------------------
void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
  for (byte I=0; I<Longueur; I++) {display.print(char(EEPROM.read(PTR))); PTR++;}; }
  
void Effacer_zone(byte X, byte Y, byte NB_caracteres) {
  byte I = 10; 
  for (byte J=0; J < I; J++)
       display.drawLine(X, Y + J, X + (NB_caracteres * 6), Y + J, BLACK);}

void CADRE_UP() {display.drawRect(0, 0, 128, 16, WHITE);}
void CADRE() {display.drawRect(0, 16, 128, 48, WHITE);}
void Affiche_cadre_du_haut(byte X) {CADRE_UP(); display.setCursor(X,4);}

