/* Programme Démonstrateur pour mettre au point l'affichage des paramètres de navigation en mode graphique.

Ce démonstrateur a été utilisé pour définir la façon dont sont représentées les données sous forme graphique
complétées par des informations numériques. Il simule une rotation permanente et constante en Lacet. Il simule
également une variation de Tangage et de Roulis entre plus et moins 90° pour vérifier la propreté de l'affichage
sachant qu'en opération réelle l'angle d'inclinaison ne devra pas dépasser les 20°. Toutefois, si la sonde se
trouvait dans une posture anormale à l'atterrissage, il serait commode d'en visualiser l'attitude sans compter
qu'en prenant en main JEKERT en démonstration on peut ainsi démontrer le comportement de la centrale
gyroscopique et en présenter les répercutions sur l'affichage qu'il soit graphique ou numérique.

>>> Pour expérimenter l'affichage du "bug" de calage gyroscopique changer la valeur de la variable
    Calage_Gyroscopique sur la ligne de code repérée avec @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@.

Le roulis toléré sera compris entre -10° et +10°.
Le Tangage toléré sera compris entre -10° et +10°.
Pour Tangage et Roulis l'affichage est limité entre -10° et +10° pour étaler le visuel.

                                                             PGM             RAM dynamique 
Programme complet non optimisé :                       10798 Octets (35%)   149 Octets (7%)
Ajout des routines d'utilisation les textes EEPROM :   10870 Octets (35%)   146 Octets (7%)
Passage de trois textes en EEPROM :                    10846 Octets (35%)   132 Octets (6%)  -24 octets
Optimisations diverses :                               10668 Octets (34%)   128 Octets (6%) -178 octets
*/

#include <EEPROM.h>

//--------------------- Déclarations pour le matériel -----------------------
#include <Adafruit_ssd1306syp.h> // Afficher graphique OLED.
// Définition des broches pour la ligne I2C.
#define broche_SDA 11 // D11 utilisée en sortie.
#define broche_SCL 12 // D12 utilisée en sortie.
// Initialisation des broches en sorties et affectations.
Adafruit_ssd1306syp display(broche_SDA, broche_SCL);
//------------------------- Variables du programme --------------------------

byte Position_aff_CAP;
byte SAVX, SAVY; // Anciennes valeurs pour effacer.
byte X,Y; // Coordonnées de l'extrémité du rayon tracé.

int Calage_Gyroscopique;
int Tangage, Roulis, Ecart_gyroscopique, Ruban;

float Alpha;

void setup()  {
  Calage_Gyroscopique = 180; // Ici constante pour tester le programme. @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  Tangage = 0, Roulis = -90;
  //------------------------------ Affiche écran initial -------------------------------
  display.initialize(); delay(300); display.clear();
  //---------------------------- Affiche la rose des CAPs ----------------------
  display.drawCircle(32, 40, 23, WHITE); // Affiche le cercle.
  //--------- Tracer les graduations tous les 30 degrés ---------
  for (int Angle = 0; Angle < 360; Angle = Angle + 30) {
       Alpha = radians(Angle);
       X = 32 + (16 * sin(Alpha)); Y = 40 - (16 * cos(Alpha));
       SAVX = 32 + (23 * sin(Alpha)); SAVY = 40 - (23 * cos(Alpha));
       display.drawLine(X, Y, SAVX, SAVY, WHITE);}
  display.drawRect(16, 0, 33, 16, WHITE); // Cadre pour l'affichage numérique du CAP.
  //------------------------------ Ruban pour le Roulis ------------------------
  display.drawRect(55, 5, 73, 6, WHITE);
  for (byte X=55; X<128; X=X+18) display.drawLine(X, 0, X, 5, WHITE);
  //----------------------------- Ruban pour le Tangage ------------------------
  display.drawRect(121, 19, 7, 45, WHITE);
  for (byte Y=19; Y<64; Y=Y+11) display.drawLine(115, Y, 120, Y, WHITE);
  //------------------- Ecriture des divers champs de données ------------------
  display.setCursor(62,18); Aff_TEXTE_EEPROM(439,3); display.setCursor(62,31); Aff_TEXTE_EEPROM(992,3);
  display.setCursor(62,43); Aff_TEXTE_EEPROM(995,4); display.print(Calage_Gyroscopique); Affiche_Degre();
  display.setCursor(56,55); Aff_TEXTE_EEPROM(999,3);  
  //------------------ Tracer le "bug" du calage gyroscopique ------------------
  Alpha = radians(Calage_Gyroscopique);
  X = 32 + (19 * sin(Alpha)); Y = 40 - (19 * cos(Alpha));
  display.drawCircle(X, Y, 2, WHITE);
  //----------------------------------------------------------------------------
  X = 33; Y = 41; // Pour ne pas effacer des pixel au premier affichage du rayon.
  //----------------------------------------------------------------------------
  display.setTextSize(1); display.update();} // Taille la plus petite.
  
void loop() {
  for (int Angle = 0; Angle < 360; Angle = Angle + 1) {
       //------------------------ Calcule les coordonnées ----------------------
       Alpha = radians(Angle);
       SAVX = X; SAVY = Y;
       X = 32 + (15 * sin(Alpha)); Y = 40 - (15 * cos(Alpha));
       //-------------------- Trace le rayon vecteur du CAP --------------------
       display.drawLine(32, 40, SAVX, SAVY, BLACK);
       display.drawLine(32, 40, X, Y, WHITE);
       //----------------- Indique la valeur numérique du CAP ------------------
       Position_aff_CAP = 22;
       if (Angle < 100) Position_aff_CAP = 25; if (Angle < 10) Position_aff_CAP = 29;
       Effacer_zone(22, 4, 4); display.setCursor(Position_aff_CAP,4);
       display.print(Angle); Affiche_Degre();
       //---------------- Affiche l'écart de route gyroscopique ----------------
       Effacer_zone(80,55,5); display.setCursor(80,55);
       Ecart_gyroscopique = Angle - Calage_Gyroscopique;
       if (Ecart_gyroscopique > 0) display.print('+');
       display.print(Ecart_gyroscopique); Affiche_Degre(); 
       //----------------------- Trace le ruban du Roulis ----------------------
       for (byte Y=6; Y<10; Y++) display.drawLine(56, Y, 126, Y, BLACK); // Efface le ruban.
       Roulis = Roulis + 2; if (Roulis > 90) Roulis = -90; Ruban = Roulis; Limite_Ruban();
       Ruban = map(Ruban,-10,+10,55,127);
       for (byte Y=6; Y<10; Y++) display.drawLine(91, Y, Ruban, Y, WHITE); // Trace le ruban.
       //-------------------- Affiche le Roulis en numérique -------------------
       Prepare_zone_aff_numerique(18);
       if (Roulis > 0) display.print('+');
       display.print(Roulis); Affiche_Degre();
       //----------------------- Trace le ruban du Tangage ---------------------
       for (byte X=122; X<127; X++) display.drawLine(X, 20, X, 62, BLACK); // Efface le ruban.
       Tangage = Tangage + 2; if (Tangage > 90) Tangage = -90; Ruban = Tangage; Limite_Ruban();
       Ruban = map(Ruban,-10,+10,63,19);
       for (byte X=122; X<127; X++) display.drawLine(X, 41, X, Ruban, WHITE); // Trace le ruban.
       //------------------- Affiche le Tangage en numérique -------------------
       Prepare_zone_aff_numerique(31);
       if (Tangage > 0) display.print('+');
       display.print(Tangage); Affiche_Degre();
       //-----------------------------------------------------------------------
  display.update();}}

void Prepare_zone_aff_numerique(byte X) {
  Effacer_zone(86,X,4); display.setCursor(86,X);}
  
void Limite_Ruban() {if (Ruban < -10) Ruban = -10; if (Ruban > 10) Ruban = 10;}
  
//------------------------- Routines de servitude --------------------------
void Aff_TEXTE_EEPROM(int PTR, byte Longueur) {
  for (byte I=0; I<Longueur; I++) {display.print(char(EEPROM.read(PTR))); PTR++;}; }
  
void Affiche_Degre() {display.print(char(247));}

void Effacer_zone(byte X, byte Y, byte NB_caracteres) {
  byte I = 10; 
  for (byte J=0; J < I; J++)
       display.drawLine(X, Y + J, X + (NB_caracteres * 6), Y + J, BLACK);}

